(* The Advanced Encryption Standard *)

(* Author:
Jose Luis Gomez Pardo
Departamento de Alxebra
Universidade de Santiago de Compostela
15782 Santiago de Compostela, Spain
e-mail: pardo@usc.es
*)

(* Copyright : 
Copyright 2003-2005 Jose Luis Gomez Pardo
Permission is hereby granted to make copies of this file for any purpose
other than direct profit, or as part of a commercial product, provided this
copyright notice is left intact. Also, permission is granted to reproduce
part or all of this file provided that the source is acknowledged.
*)

(* Version: 2.1 *)

(* Keywords: Cryptography, block cipher, Rijndael, AES, authentication, OMAC *)

(* Requirements: Mathematica  v. 4.0 or higher *)

If[$VersionNumber < 4.0, Print[
  "This package requires Mathematica version 4.0 or higher."]];
  
(* Summary:
This package contains a Mathematica implementation of the Advanced
Encryption Standard (AES) in ECB, CBC and CFB128 modes, as well as 
several variants of AES-based CBC-MAC authentication, including OMAC.
*)

(* Discussion: 
See the comments included in the accompanying notebook "Rijndael.nb" 
*)

BeginPackage["Cryptography`AES`", "Algebra`FiniteFields`"];

TextToKey::usage = "TextToKey[text,length] generates a key of the specified
length (128, 192, or 256-bit) from a text string.";

RandomKey::usage = "RandomKey[length] generates a random key of the given
length (128, 192, or 256-bit).";

AesTextEncrypt::usage = "AesTextEncrypt[plaintext, key, (mode)] encrypts a
plaintext string using the given key and mode. The latter can be: ecb, cbc
or cfb; if no mode is specified, cbc is used by default.";

AesTextDecrypt::usage = "AesTextDecrypt[ciphertext, key, (mode)] decrypts a
ciphertext string using the given key and mode. The latter can be: ecb, cbc
or cfb; if no mode is specified, cbc is used by default.";

AesBinaryEncrypt::usage = "AesBinaryEncrypt[filename, cfilename, key,(mode)]
reads the file named filename from disk, encrypts it with key in the
specified mode (ecb, cbc or cfb), and writes it to cfilename. If the
optional argument mode is omitted, then cbc mode is used by default. This
function checks if there is a file named cfilename in the working directory
and, in that case, the user is prompted about whether this file should be
overwritten or a different name should be used. This check can be overridden
by setting the option CheckFile to False.";

AesBinaryDecrypt::usage = "AesBinaryDecrypt[cfilename, filename, key,(mode)]
reads the file named cfilename from disk, decrypts it with key in the
specified mode (ecb, cbc or cfb), and writes it to filename. If the optional
argument mode is omitted, then cbc mode is used by default. This function
checks if there is a file named filename in the working directory and,
in that case, the user is prompted about whether this file should be
overwritten or a different name should be used. This check can be overridden
by setting the option CheckFile to False.";

CheckFile::usage = "CheckFile is an option to AesBinaryEncrypt and
AesBinaryDecrypt. By default, these functions check whether a file with the
specified name exists on disk. If set to False, the file is written to disk
without checking.";

StringToFile::usage = "StringToFile[str, filename] writes the string str to
a file named filename, in binary form.";

FileToString::usage = "FileToString[filename] reads the file named filename
in binary form and converts it to a string.";

BytesToHexString::usage = "BytesToHexString[bytes] converts a list of bytes
to a hex string.";

HexStringToBytes::usage = "HexStringToBytes[str] converts a hex string to
the corresponding list of bytes.";

KeyStringToBytes::usage = "KeyStringToBytes[keystring] converts an AES key
given as a hex string to a list of bytes.";

RandomKeyString::usage = "RandomKeyString[length] generates a
(pseudo-)random key of the specified length (128, 192, or 256-bit) in the
form of a hex string.";

TextToKeyString::usage = "TextToKeyString[text, length] generates a key of
the specified length (128, 192, or 256-bit), given as a hex string, from
the supplied text.";

AesCbcMac::usage = "AesCbcMac[filename, keystring] uses the secret key to
compute the MAC associated to the file named filename. The output is an
8-byte hex string.";

AesCbcEmac::usage = "AesCbcEmac[filename, keystring1, keystring2] uses the
secret keystring1 and keystring2 to compute the EMAC associated to the file
named filename. The output is a 16-byte hex string.";

AesOmac::usage = "AesOmac[filename, keystring, (type)] produces the OMAC
for the given file, using keystring as key. The type should be given as 1
for OMAC1 and 2 for OMAC2. If no type is specified, then OMAC1 is used by
default.";

AesTextOmac::usage = "AesTextOmac[text, keystring, (type)] produces the OMAC
for the given text, using keystring as key. The type should be given as 1
for OMAC1 and 2 for OMAC2. If no type is specified, then OMAC1 is used by
default.";

DisplayRijndaelObject::usage = "DisplayRijndaelObject[objectname] displays
the corresponding Rijndael constant or table (objectname can be Sbox, ISbox,
MC, IMC or Rcon).";

VarTextTest::usage = "VarTextTest[length] computes the input of the variable
text tests for the given keylength, and produces as output the result of
encrypting these texts.";

VarKeyTest::usage = "VarKeyTest[length] computes the input of the variable
key tests for the given keylength, and produces as output the corresponding
ciphertexts.";

AesEncryptTest::usage = "AesEncryptTest[plaintext, keystring, (iv), mode]
encrypts plaintext in the specified mode, with the supplied key and iv (all
of them given as hex strings). If no iv is supplied, then the string
corresponding to the first 16 bytes in ascending order is used by default
(except in ECB mode, where it is not needed).";

AesDecryptTest::usage = "AesDecryptTest[ciphertext, keystring, (iv), mode]
decrypts ciphertext in the specified mode, with the supplied key and iv (all
of them given as hex strings). If no iv is supplied, then the string
corresponding to the first 16 bytes in ascending order is used by default
(except in ECB mode, where it is not needed).";

AesRandomTest::usage = "AesRandomTest[n, l] generates n random text strings of
length at most l and n random AES keys which are used to encrypt and decrypt
these text strings. For each of these tests, the encryption-decryption mode
is also chosen at random. If all the results match, the test is declared
successfully passed.";

AesRandomBinaryTest::usage = "AesRandomBinaryTest[filename, (mode), n] submits
the file named filename to n successive encryption-decryption processes in
the given mode with n randomly chosen keys. If no mode is specified, then
cbc is used by default. The test is successful if the final result is
identical to the original file.";

OmacTest::usage = "OmacTest[text, keystring, (type)] takes as input a
hexadecimal text string, a hexadecimal keystring and the OMAC type (either 1
or 2) and produces as output the corresponding MAC, also in the form of a
hex string. If no type is specified, OMAC1 is used by default.";

TextToKey::noemp = "The empty text string cannot be used to generate a
key.";

TextToKey::noleng = "The key length must be equal to 128, 192, or 256.";

RandomKey::noleng = "The key length must be equal to 128, 192, or 256.";

CheckKey::nobytes = "The supplied key `` is not a list of bytes.";

CheckKey::noleng = "The supplied key `` does not have valid length.";

AesTextDecrypt::notaes = "`` is not an AES-encrypted text.";

AesTextDecrypt::badbyte = "Error, probably due to using a wrong key or
a wrong mode.";

AesBinaryDecrypt::notaes = "`` is not an AES-encrypted file.";

AesBinaryDecrypt::badbyte = "Error, probably due to using a wrong key
or a wrong mode.";

HexStringToBytes::nohex = "`` is not a valid hex string.";

KeyStringToBytes::noleng = "The key `` does not have valid length.";

AesOmac::notype = "The OMAC type must be either 1 or 2.";

DisplayRijndaelObject::noobj = "The supplied argument is not a valid name
of a Rijndael object.";

AesEncryptTest::noleng = "The IV `` does not have valid length.";

AesDecryptTest::noleng = "The IV `` does not have valid length.";

OmacTest::notype = "The OMAC type must be either 1 or 2.";

Begin["`Private`"];

Scan[SetAttributes[#, Listable] &, {ByteTimes2, InvByte, AffMap, ByteMult,
SubBytes, InvSubBytes, ToHexDigit, ByteToHex, FromCode}];

ir={1,1,0,1,1,0,0,0,1};
F = GF[2,ir];

ByteToFieldElement[byte_Integer] := F[Reverse[IntegerDigits[byte, 2, 8]]];

FieldElementToByte[element_] := FromDigits[Reverse[element[[1]]], 2];

ByteTimes2[byte_Integer] := If[byte < 128, 2 byte, BitXor[283, 2 byte]];

MultTable := 
    Module[{mt = Table[Range[256] - 1, {15}], 
        n = NestList[ByteTimes2, Range[256] - 1, 3]},
      Do[mt[[2^i]] = n[[i + 1]], {i, 1, 3}];
      Do[mt[[j]] = BitXor[mt[[2^i]], mt[[j - 2^i]]], {i, 1, 3}, {j, 2^i + 1, 
          2^(i + 1) - 1}];
            mt];

InvByte[byte_Integer] := If[byte == 0, 0, FieldElementToByte[1/ByteToFieldElement[byte]]];

vect = {1, 1, 0, 0, 0, 1, 1, 0};
M = Table[RotateRight[{1, 0, 0, 0, 1, 1, 1, 1}, i], {i, 0, 7}];
IM = Inverse[M, Modulus -> 2];

AffMap[byte_Integer] := 
    Module[{coeffs = Reverse[IntegerDigits[byte, 2, 8]]}, 
      FromDigits[Reverse[Mod[M.coeffs + vect, 2]], 2]];

S = Module[{x = Range[2, 255], s = Range[0, 255], a, b}, 
      Do[a = First[x]; b = InvByte[a]; s[[a + 1]] = b; s[[b + 1]] = a; 
        x = Complement[x, {a, b}], {127}]; AffMap[s]];

IS = Module[{is = Range[256]}, 
      Do[is[[i]] = Flatten[Position[S, i - 1]][[1]] - 1, {i, 256}]; is];

MC = Table[RotateRight[{2, 3, 1, 1}, i], {i, 0, 3}];

IMC = Table[RotateRight[{14, 11, 13, 9}, i], {i, 0, 3}];

MT = MultTable;

ByteMult[byte1_Integer, byte2_Integer] := MT[[byte1]][[byte2 + 1]];

Rcon = NestList[ByteMult[2, #] &, {1, 0, 0, 0}, 13];

SubBytes[byte_Integer] := S[[byte + 1]];

InvSubBytes[byte_Integer] := IS[[byte + 1]];

ShiftLeft[mat_List] := 
    Module[{m = mat}, Do[m[[i]] = RotateLeft[m[[i]], i - 1], {i, 2, 4}]; m];

ShiftRows[states_List] := ShiftLeft[#] & /@ states

ShiftRight[mat_List] := 
    Module[{m = mat}, Do[m[[i]] = RotateRight[m[[i]], i - 1], {i, 2, 4}]; m];

InvShiftRows[states_List] := ShiftRight[#] & /@ states

MixColumns[states_List] := Inner[ByteMult, MC, #, BitXor] & /@ states

InvMixColumns[states_List] := Inner[ByteMult, IMC, #, BitXor] & /@ states

KeyExpansion[key_List] :=
    Module[{l = Length[key], k = Partition[key, 4], s, w, temp},
      s = l + 28; w = Table[{0, 0, 0, 0}, {s}]; Nk = l/4; 
      Do[w[[i]] = k[[i]], {i, Nk}]; 
      Do[temp = w[[i - 1]]; 
        If[ Mod[i - 1, Nk] == 0, 
          temp = BitXor[SubBytes[RotateLeft[temp]], Rcon[[(i - 1)/Nk]]], 
          If[Nk == 8 && Mod[i - 5, Nk] == 0, temp = SubBytes[temp]]]; 
        w[[i]] = BitXor[w[[i - Nk]], temp], {i, Nk + 1, s}];
      Transpose[#] & /@ Partition[w, 4]];

AddRoundKey[key_List, states_List] := BitXor[key, #] & /@ states

BytesToStates[bytes_List] := 
    Transpose[Partition[#, 4]] & /@ Partition[bytes, 16];

StatesToBytes[states_List] := Flatten[Transpose[#] & /@ states];

RijndaelRound[states_List, roundk_List] := 
    AddRoundKey[roundk, MixColumns[ShiftRows[SubBytes[states]]]];

RijndaelEncrypt[bytes_List, key_List] := 
    Module[{ek = KeyExpansion[key], s = BytesToStates[bytes], 
        Nr = Length[key]/4 + 6},
      s = AddRoundKey[ek[[1]], s];
      s = Fold[RijndaelRound[#1, #2] &, s, Drop[Rest[ek], -1]];
      s = AddRoundKey[ek[[Nr + 1]], ShiftRows[SubBytes[s]]];
             StatesToBytes[s]];

RijndaelInvRound[states_List, roundk_List] := 
    InvSubBytes[InvShiftRows[InvMixColumns[AddRoundKey[roundk, states]]]];

RijndaelDecrypt[bytes_List, key_List] := 
    Module[{ek = KeyExpansion[key], s = BytesToStates[bytes], 
        Nr = Length[key]/4 + 6},
      s = InvSubBytes[InvShiftRows[AddRoundKey[ek[[Nr + 1]], s]]];
      s = Fold[RijndaelInvRound[#1, #2] &, s, Reverse[Drop[Rest[ek], -1]]];
      s = AddRoundKey[ek[[1]], s];
             StatesToBytes[s]];

CbcEncrypt[bytes_List, key_List, iv_List] := 
    Module[{b = Partition[Join[iv, bytes], 16], k = key},
      Flatten[
        FoldList[RijndaelEncrypt[BitXor[#1, #2], k] &, First[b], Rest[b]]]];

CbcDecrypt[bytes_List, key_List] :=
    Module[{b = Partition[bytes, 16], k = key}, 
      Flatten[MapThread[
          BitXor[#1, RijndaelDecrypt[#2, k]] &, {Drop[b, -1], Rest[b]}]]];

CfbEncrypt[bytes_List, key_List, iv_List] := 
    Module[{b = Partition[Join[iv, bytes], 16], k = key},
      Flatten[
        FoldList[BitXor[RijndaelEncrypt[#1, k], #2] &, First[b], Rest[b]]]];

CfbDecrypt[bytes_List, key_List] :=
    Module[{b = Partition[bytes, 16], k = key}, 
      Flatten[MapThread[
          BitXor[RijndaelEncrypt[#1, k], #2] &, {Drop[b, -1], Rest[b]}]]];

RijndaelEncryptMode[bytes_List, key_List, mode_String, iv_List] := 
    Module[{b = bytes, k = key, m = mode, i = iv}, 
      Which[m == "cbc", CbcEncrypt[b, k, i],
                    m == "ecb", RijndaelEncrypt[b, k],            
                    True, CfbEncrypt[b, k, i]]];

RijndaelDecryptMode[bytes_List, key_List, mode_String] := 
    Module[{b = bytes, k = key, m = mode}, 
      Which[m == "cbc", CbcDecrypt[b, k],
                    m == "ecb", RijndaelDecrypt[b, k],            
                    True, CfbDecrypt[b, k]]];

PadByteList[bytes_?(VectorQ[#, IntegerQ] &)] := Module[{b = bytes, l, rem},
      l = Length[b]; rem = Mod[l, 16]; 
      PadRight[b, l + 16 - rem, 
        Join[Table[Random[Integer, 255], {15}], {16 - rem}]]];

UnpadByteList[bytes_?(VectorQ[#, IntegerQ] &)] := 
    Module[{b = bytes, l = Last[bytes]}, 
      If [l > 16 || l == 0, Return[], b = Drop[b, -l]]; b];

FileToByteList[filename_String] := Module[{b}, 
		If[$VersionNumber >= 5.1 , b = BinaryReadList[filename], 
		  b = Experimental`BinaryImport[filename, {"Byte" ..}]; 
       If[ b == {EndOfFile}, b = {}]]; b];

ByteListToFile[bytes_?(VectorQ[#, IntegerQ] &), filename_String] :=
    Module[{b = bytes}, 
    If[$VersionNumber >= 5.1, BinaryWrite[filename, b, "Byte"]; Close[filename],
        If[b == {}, OpenWrite[filename]; Close[filename],
          Experimental`BinaryExport[filename, b, {"Byte" ..}]]]];

TextToKey[text_String, length_] := 
    Module[{t = text, l = length}, 
      If[StringLength[t] == 0, Message[TextToKey::noemp]; Return[]]; 
      If[MemberQ[{128, 192, 256}, l],
        While[StringLength[t] < l/8, t = StringJoin[t, t]]; 
        ToCharacterCode[StringTake[t, l/8]], Message[TextToKey::noleng]; 
        Return[]]];

RandomKey[length_] := 
    If[MemberQ[{128, 192, 256}, length], 
      Table[Random[Integer, 255], {length/8}],
      Message[RandomKey::noleng]; Return[]];

CheckKey[key_List] := 
    Module[{k = key}, 
      If[! VectorQ[k, (0 <= # <= 255) &], Message[CheckKey::nobytes, key]; 
        Return[]];
      If[! MemberQ[{16, 24, 32}, Length[k]], Message[CheckKey::noleng, key]; 
        Return[]]; k];

CheckMode[mode_String] := 
    Module[{prompt, m = mode}, While[m != "cbc" && m != "ecb" && m != "cfb",
        prompt = StringJoin["Mode ", m, " is not valid. Please input: \r\n
      cbc  for CBC mode \r\n
      ecb  for ECB mode \r\n
      cfb  for CFB mode \r\n
      a  to abort"]; 
         m = InputString[prompt];
          If[m == "a", Return[]]]; m];

AesTextEncrypt[plaintext_String, key_List, mode_String:"cbc"] :=
    Module[{b = ToCharacterCode[plaintext], k = CheckKey[key], 
        m = CheckMode[mode], iv = Table[Random[Integer, 255], {16}]},
      If[k === Null, Return[]];
      If[m === Null, Print["Encryption aborted by user"]; Return[]];
      FromCharacterCode[RijndaelEncryptMode[PadByteList[b], k, m, iv]]];

AesTextDecrypt[ciphertext_String, key_List, mode_String:"cbc"] := 
    Module[{c = ToCharacterCode[ciphertext], k = CheckKey[key], 
        m = CheckMode[mode]},
      If[k === Null, Return[]];
      If[m === Null, Print["Decryption aborted by user"]; Return[]];
      If[ Mod[Length[c], 16] != 0, 
        Message[AesTextDecrypt::notaes, ciphertext]; Return[]];
      c = UnpadByteList[RijndaelDecryptMode[c, k, m]];
      If[c === Null, Message[AesTextDecrypt::badbyte]; Return[], 
        FromCharacterCode[c]]];

CheckFileName[filename_String] :=
    Module[{answer, prompt, f = filename},
      While[FileNames[f] != {},
        prompt = 
          StringJoin["A file named ", f, 
            " already exists. Please input: \r\n 
o  to overwrite \r\n 
c  to choose another file name \r\n 
anything else to abort"];
        answer = InputString[prompt];
        Which[
           answer == "o", Break[],
           answer == "c", prompt = "Enter new file name:"; 
          f = InputString[prompt],
           True, Return[]]]; f];

AesBinaryEncrypt[filename_String, cfilename_String, key_List, 
      mode_String:"cbc", opts___Rule] :=
    Module[{b = FileToByteList[filename], cf = cfilename, k = CheckKey[key], 
        m = CheckMode[mode], iv = Table[Random[Integer, 255], {16}], 
        ch = CheckFile /. {opts} /. Options[AesBinaryEncrypt]},
      If[k === Null || b === $Failed, Return[]];
      If[m === Null, Print["Encryption aborted by user"]; Return[]];
      If[ch, cf = CheckFileName[cf]; 
        If[cf === Null, Print["Encryption aborted by user"]; Return[]]];
      ByteListToFile[RijndaelEncryptMode[PadByteList[b], k, m, iv], cf]];

Options[AesBinaryEncrypt] = {CheckFile -> True};

AesBinaryDecrypt[cfilename_String, filename_String, key_List, 
      mode_String:"cbc", opts___Rule] := 
    Module[{b = FileToByteList[cfilename], f = filename, k = CheckKey[key], 
        m = CheckMode[mode], 
        ch = CheckFile /. {opts} /. Options[AesBinaryDecrypt]},
      If[k === Null || b === $Failed, Return[]];
      If[ Mod[Length[b], 16] != 0, 
        Message[AesBinaryDecrypt::notaes, cfilename]; Return[]];
      If[m === Null, Print["Decryption aborted by user"]; Return[]];
      If[ch, f = CheckFileName[f]; 
        If[f === Null, Print["Decryption aborted by user"]; Return[]]];
      b = UnpadByteList[RijndaelDecryptMode[b, k, m]];
      If[b === Null, Message[AesBinaryDecrypt::badbyte]; Return[]]; 
      ByteListToFile[b, f]];

Options[AesBinaryDecrypt] = {CheckFile -> True};

StringToFile[str_String, filename_String] := 
    Module[{f = CheckFileName[filename]}, 
      If[f === Null, Print["Operation aborted by user"]; Return[], 
        ByteListToFile[ToCharacterCode[str], f]]];

FileToString[filename_String] := 
    Module[{b = FileToByteList[filename]}, 
      If[b === $Failed, Return[], FromCharacterCode[b]]];

ToHexDigit[n_Integer?(0 <= # <= 15 &)] := 
    If[n <= 9, ToString[n], FromCharacterCode[n + 87]];
    
ByteToHex[byte_Integer?(0 <= # <= 255 &)] := 
    StringJoin[ToHexDigit[IntegerDigits[byte, 16, 2]]];

BytesToHexString[bytes_?(VectorQ[#, (0 <= # <= 255) &] &)] := 
    StringJoin[ByteToHex[bytes]];

FromCode[n_Integer] := If[n <= 57, n - 48, n - 87];

HexStringToBytes[str_String] := 
    Module[{b = 
          FromDigits[#, 16] & /@ 
            Partition[FromCode[ToCharacterCode[ToLowerCase[str]]], 2]},
      If[! VectorQ[b, (0 <= # <= 255) &], 
        Message[HexStringToBytes::nohex, str]; Return[], b]];

KeyStringToBytes[keystring_String] := 
    Module[{k = HexStringToBytes[keystring]},
      If[k === Null, Return[]];
      If[! MemberQ[{16, 24, 32}, Length[k]], 
        Message[KeyStringToBytes::noleng, keystring]; Return[]];
      k];

RandomKeyString[length_] := 
    Module[{k = RandomKey[length]}, 
      If[k === Null, Return[], BytesToHexString[k]]];

TextToKeyString[text_String, length_] :=
    Module[{k = TextToKey[text, length]}, 
      If[k === Null, Return[], BytesToHexString[k]]];

RijndaelBasicCbcMac[bytes_List, key_List] := 
    Module[{b = Join[bytes, {128}], l, iv = Table[0, {16}]},
      l = Length[b]; 
      If[ Mod[l, 16] != 0, b = PadRight[b, l + 16 - Mod[l, 16]]];
          b = Partition[Join[iv, b], 16];      
      Fold[RijndaelEncrypt[BitXor[#1, #2], key] &, First[b], Rest[b]]];

AesCbcMac[filename_String,  keystring_String] := 
    Module[{k = KeyStringToBytes[keystring], b = FileToByteList[filename], 
        m},
      If[k === Null || b === $Failed, Return[]];
      m = RijndaelBasicCbcMac[b, k];
      BytesToHexString[Take[m, 8]]];

AesCbcEmac[filename_String, keystring1_String, keystring2_String] := 
    Module[{k1 = KeyStringToBytes[keystring1], 
        k2 = KeyStringToBytes[keystring2], b = FileToByteList[filename], m},
      If[k1 === Null || k2 === Null || b === $Failed, Return[]];
      m = RijndaelBasicCbcMac[b, k1];
      BytesToHexString[RijndaelEncrypt[m, k2]]];

BitRotateLeft[bytes_List] := 
    FromDigits[#, 2] & /@ 
      Partition[RotateLeft[Flatten[IntegerDigits[bytes, 2, 8]]], 8];

BitRotateRight[bytes_List] := 
    FromDigits[#, 2] & /@ 
      Partition[RotateRight[Flatten[IntegerDigits[bytes, 2, 8]]], 8];

MultBy2[bytes_List] := 
    Module[{f = First[bytes], r = BitRotateLeft[bytes]}, 
      If[f < 128, r, BitXor[r, Join[Table[0, {15}], {134}]]]];

DivBy2[bytes_List] := Module[{l = Last[bytes], r = BitRotateRight[bytes]},
      If[EvenQ[l], r, BitXor[r, Join[Table[0, {15}], {67}]]]];

RijndaelOmac[bytes_List, key_List, type_Integer] :=
    Module[{b = bytes, k = key, fullblock = False, iv = Table[0, {16}], l, lb, x},
      l = Length[b]; 
      If[ Mod[l, 16] != 0 || l == 0, b = Join[b, {128}]; 
        b = PadRight[b, l + 16 - Mod[l, 16]], fullblock = True]; 
      b = Partition[Join[iv, b], 16]; lb = Last[b];  
      x = RijndaelEncrypt[iv, k];
      If[fullblock, b = ReplacePart[b, BitXor[lb, MultBy2[x]], -1], 
        If[type == 1, 
          b = ReplacePart[b, BitXor[lb, MultBy2[MultBy2[x]]], -1], 
          b = ReplacePart[b, BitXor[lb, DivBy2[x]] , -1]]];
      Fold[RijndaelEncrypt[BitXor[#1, #2], k] &, First[b], Rest[b]]];

AesOmac[filename_String, keystring_String, type_:1] := 
    Module[{b = FileToByteList[filename], k = KeyStringToBytes[keystring]},
      If[k === Null || b === $Failed, Return[]];
      If[! MemberQ[{1, 2}, type], Message[AesOmac::notype]; Return[]];
      BytesToHexString[RijndaelOmac[b, k, type]]];

AesTextOmac[text_String, keystring_String, type_:1] := 
    Module[{b = ToCharacterCode[text], k = KeyStringToBytes[keystring]},
      If[k === Null, Return[]];
      If[! MemberQ[{1, 2}, type], Message[AesOmac::notype]; Return[]];
      BytesToHexString[RijndaelOmac[b, k, type]]];

DisplayRijndaelObject[objectname_String] :=
    Which[
      objectname == "Sbox",
        TableForm[ByteToHex[Partition[S, 16]], 
        TableHeadings -> ToHexDigit[{Range[16] - 1, Range[16] - 1}], 
        TableSpacing -> {1, 1}],
      objectname == "ISbox",
        TableForm[ByteToHex[Partition[IS, 16]], 
        TableHeadings -> ToHexDigit[{Range[16] - 1, Range[16] - 1}], 
        TableSpacing -> {1, 1}],
      objectname == "MC",
         MatrixForm[ByteToHex[MC]],
      objectname == "IMC",
         MatrixForm[ByteToHex[IMC]],
      objectname == "Rcon",
         TableForm[ByteToHex[Rcon], TableHeadings -> {Automatic, None}],
      True,
         Message[DisplayRijndaelObject::noobj]
      ];

VarTextTest[length_?(MemberQ[{128, 192, 256}, #] &)] := 
    Module[{k = Table[0, {length/8}], 
        t = FromDigits[#, 2] & /@ 
            Partition[
              Flatten[NestList[ReplacePart[RotateRight[#], 1, 1] &, 
                  Join[{1}, Table[0, {127}]], 127]], 8]}, 
      t = RijndaelEncrypt[t, k];
      Do[Print[i, ": ", BytesToHexString[Partition[t, 16][[i]]]], {i, 128}]];

VarKeyTest[length_?(MemberQ[{128, 192, 256}, #] &)] := 
    Module[{bkl = 
          Partition[#, 8] & /@ 
            NestList[ReplacePart[RotateRight[#], 1, 1] &, 
              Join[{1}, Table[0, {length - 1}]], length - 1], 
        pt = Table[0, {16}]},
      Do[Print[i, ": ", 
          BytesToHexString[
            RijndaelEncrypt[pt, FromDigits[#, 2] & /@ bkl[[i]]]]], {i, 
          length}]];

AesEncryptTest[plaintext_String, keystring_String, mode_String, 
      iv_String:BytesToHexString[Range[16] - 1]] :=
    Module[{b = HexStringToBytes[plaintext], k = KeyStringToBytes[keystring], 
        i = HexStringToBytes[iv], m = CheckMode[mode]},
      If[k === Null || i === Null, Return[]];
      If[m === Null, Print["Test aborted by user"]; Return[]];
      If[Length[i] != 16, Message[AesEncryptTest::noleng, iv]; Return[]];
      b = BytesToHexString[RijndaelEncryptMode[b, k, m, i]]; 
      If[m == "ecb", b, StringDrop[b, 32]]];

AesDecryptTest[ciphertext_String, keystring_String, mode_String, 
      iv_String:BytesToHexString[Range[16] - 1]] :=
    Module[{b = HexStringToBytes[ciphertext], k = KeyStringToBytes[keystring],
         i = HexStringToBytes[iv], m = CheckMode[mode]},
      If[k === Null || i === Null, Return[]];
      If[m === Null, Print["Test aborted by user"]; Return[]];
      If[Length[i] != 16, Message[AesDecryptTest::noleng, iv]; Return[]];
      If[m != "ecb", b = Join[i, b]];
      BytesToHexString[RijndaelDecryptMode[b, k, m]]];

AesRandomTest[n_Integer?Positive, l_Integer?Positive] := 
    Module[{t, k, j, i = 1, m = {"ecb", "cbc", "cfb"}, cond = True}, 
      While[cond && i <= n, 
        k = RandomKey[64 Random[Integer, {2, 4}]]; 
        t = FromCharacterCode[
            Table[Random[Integer, {32, 255}], {Random[Integer, {1, l}]}]];
        j = Random[Integer, {1, 3}];
        cond = (AesTextDecrypt[AesTextEncrypt[t, k, m[[j]]], k, m[[j]]] == t);
         If[Mod[i, 100] == 0, Print[i, " tests"]]; i++]; 
      If[cond, Print[n, " tests successfully passed"], 
        Print["Test failed at plaintext: \r\n", t, "\r\n with key: \r\n", k, 
          "\r\n in ", ToUpperCase[m[[j]]], " mode."]]];

AesRandomBinaryTest[filename_String, mode_String:"cbc", n_Integer?Positive] :=
     Module[{b = FileToByteList[filename], c, k, m = CheckMode[mode]},
      If[b === $Failed, Return[]];
      If[m === Null, Print["Test aborted by user"]; Return[]]; 
       Do[k = RandomKey[64 Random[Integer, {2, 4}]];
        AesBinaryEncrypt[filename, filename, k, m, CheckFile -> False];
        AesBinaryDecrypt[filename, filename, k, m, CheckFile -> False], {n}]; 
      c = FileToByteList[filename];
      If[b == c,
        Print["Test successfully passed"], Print["Test failed"]]];

OmacTest[text_String, keystring_String, type_:1] := 
    Module[{b = HexStringToBytes[text], k = KeyStringToBytes[keystring]},
      If[b === Null || k === Null, Return[]];
      If[! MemberQ[{1, 2}, type], Message[OmacTest::notype]; Return[]];
      BytesToHexString[RijndaelOmac[b, k, type]]];

End[];

EndPackage[];